package gs.timing.v0

import cats.effect.Sync
import cats.syntax.all.*

/** Primary entrypoint for the `gs-timing` library. This class is an interface
  * for working with monotonic time:
  *
  *   - Getting the current value of some monotonic clock.
  *   - Calculating the elapsed time between two events.
  *
  * This class/library is not useful for anything related to human date/time
  * operations, and is focused on precise machine-level timing.
  *
  * ## Sharing Instances
  *
  * System instances of this class can be shared - one instance can be used
  * across an application.
  *
  * Manual instances of this class are user-controlled and should be shared with
  * caution, and will not necessarily give consistent results across concurrent
  * calls (depending on what each call does). Since these are typically reserved
  * for testing purposes, creating an instance per test is advised.
  *
  * @param monotonicProvider
  *   The [[MonotonicProvider]], which determines how nanoseconds are
  *   calculated.
  */
final class Timing[F[_]: Sync](
  monotonicProvider: MonotonicProvider[F]
):
  /** @return
    *   The current monotonic time (expressed as nanoseconds).
    */
  def monotonic(): F[Long] = monotonicProvider.monotonic()

  /** @return
    *   A new [[MonotonicTimer]] based on the current monotonic time.
    */
  def start(): F[MonotonicTimer[F]] =
    monotonic().map(start => new MonotonicTimer[F](start, monotonic))

object Timing:

  /** @return
    *   A new [[Timing]] instance based on the system's underlying monotonic
    *   clock. This implementation delegates to `System.nanoTime()`.
    */
  def system[F[_]: Sync]: Timing[F] =
    new Timing[F](MonotonicProvider.system[F])

  /** @return
    *   A new [[Timing]] instance, always initialized to 0, that is manually
    *   controlled by the user.
    */
  def manual[F[_]: Sync]
    : F[(MonotonicProvider.ManualTickProvider[F], Timing[F])] =
    MonotonicProvider.manual[F].map(p => p -> new Timing[F](p))

end Timing
